@@ -33,7 +33,7 @@
 #define FCGID_APPLY_TRY_COUNT 2
 #define FCGID_REQUEST_COUNT 32
 
-static fcgid_procnode *apply_free_procnode(server_rec * s,
+static fcgid_procnode *apply_free_procnode(request_rec *r,
                                            fcgid_command * command)
 {
     /* Scan idle list, find a node match inode, deviceid and groupid
@@ -51,7 +51,7 @@ static fcgid_procnode *apply_free_procnode(server_rec * s,
     previous_node = proctable_get_idle_list();
     busy_list_header = proctable_get_busy_list();
 
-    safe_lock(s);
+    proctable_lock(r);
     current_node = &proc_table[previous_node->next_index];
     while (current_node != proc_table) {
         next_node = &proc_table[current_node->next_index];
@@ -68,22 +68,23 @@ static fcgid_procnode *apply_free_procnode(server_rec * s,
             current_node->next_index = busy_list_header->next_index;
             busy_list_header->next_index = current_node - proc_table;
 
-            safe_unlock(s);
+            proctable_unlock(r);
             return current_node;
-        } else
+        }
+        else
             previous_node = current_node;
 
         current_node = next_node;
     }
-    safe_unlock(s);
+    proctable_unlock(r);
 
     /* Found nothing */
     return NULL;
 }
 
 static void
-return_procnode(server_rec * s,
-                fcgid_procnode * procnode, int communicate_error)
+return_procnode(request_rec *r,
+                fcgid_procnode *procnode, int communicate_error)
 {
     fcgid_procnode *previous_node, *current_node, *next_node;
     fcgid_procnode *proc_table = proctable_get_table_array();
@@ -91,7 +92,7 @@ return_procnode(server_rec * s,
     fcgid_procnode *idle_list_header = proctable_get_idle_list();
     fcgid_procnode *busy_list_header = proctable_get_busy_list();
 
-    safe_lock(s);
+    proctable_lock(r);
 
     /* Unlink the node from busy list first */
     previous_node = busy_list_header;
@@ -102,7 +103,8 @@ return_procnode(server_rec * s,
             /* Unlink from busy list */
             previous_node->next_index = current_node->next_index;
             break;
-        } else
+        }
+        else
             previous_node = current_node;
         current_node = next_node;
     }
@@ -112,24 +114,24 @@ return_procnode(server_rec * s,
         /* Link to error list */
         procnode->next_index = error_list_header->next_index;
         error_list_header->next_index = procnode - proc_table;
-    } else {
+    }
+    else {
         /* Link to idle list */
         procnode->next_index = idle_list_header->next_index;
         idle_list_header->next_index = procnode - proc_table;
     }
 
-    safe_unlock(s);
+    proctable_unlock(r);
 }
 
-static int
-count_busy_processes(server_rec * s, fcgid_command * command)
+static int count_busy_processes(request_rec *r, fcgid_command *command)
 {
     int result = 0;
     fcgid_procnode *previous_node, *current_node, *next_node;
     fcgid_procnode *proc_table = proctable_get_table_array();
     fcgid_procnode *busy_list_header = proctable_get_busy_list();
 
-    safe_lock(s);
+    proctable_lock(r);
 
     previous_node = busy_list_header;
     current_node = &proc_table[previous_node->next_index];
@@ -146,7 +148,7 @@ count_busy_processes(server_rec * s, fcgid_command * command)
         current_node = next_node;
     }
 
-    safe_unlock(s);
+    proctable_unlock(r);
 
     return result;
 }
@@ -160,9 +162,6 @@ apr_status_t bucket_ctx_cleanup(void *thectx)
      */
     fcgid_bucket_ctx *ctx = (fcgid_bucket_ctx *) thectx;
     request_rec *r = ctx->ipc.request;
-    server_rec *s = r->server;
-    fcgid_server_conf *sconf = ap_get_module_config(s->module_config,
-                                                    &fcgid_module);
 
     /* Free bucket buffer */
     if (ctx->buffer) {
@@ -173,37 +172,31 @@ apr_status_t bucket_ctx_cleanup(void *thectx)
     proc_close_ipc(&ctx->ipc);
 
     if (ctx->procnode) {
-        /* FIXME See BZ #47483 */
+        ++ctx->procnode->requests_handled;
+
         /* Return procnode
-           I will return this slot to idle(or error) list except:
-           I take too much time on this request( greater than BusyTimeout) ),
-           so the process manager may have put this slot from busy list to error
-           list, and the contain of this slot may have been modified
-           In this case I will do nothing and return, let the process manager 
-           do the job   
+           I will return this slot to idle(or error) list
          */
-        int dt = (int)
-            (apr_time_sec(apr_time_now()) - apr_time_sec(ctx->active_time));
-        if (dt > sconf->busy_timeout) { /* can't reference from procnode->cmdopts
-                                         * because procnode could have been reused
-                                         */
-            /* Do nothing but print log */
+        if (ctx->procnode->diewhy == FCGID_DIE_BUSY_TIMEOUT) {
             ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
-                          "mod_fcgid: process busy timeout (%d), took %d seconds for this request",
-                          sconf->busy_timeout, dt);
-        } else if (ctx->has_error) {
+                          "mod_fcgid: %s took longer than busy timeout "
+                          "(%d secs)",
+                          r->uri,
+                          ctx->procnode->cmdopts.busy_timeout);
+            return_procnode(r, ctx->procnode, 1 /* busy timeout */ );
+        }
+        else if (ctx->has_error) {
             ctx->procnode->diewhy = FCGID_DIE_COMM_ERROR;
-            return_procnode(s, ctx->procnode,
-                            1 /* communication error */ );
-        } else if (ctx->procnode->cmdopts.max_requests_per_process
-                   && ++ctx->procnode->requests_handled >=
-                   ctx->procnode->cmdopts.max_requests_per_process) {
+            return_procnode(r, ctx->procnode, 1 /* communication error */ );
+        }
+        else if (ctx->procnode->cmdopts.max_requests_per_process
+                 && ctx->procnode->requests_handled >=
+                 ctx->procnode->cmdopts.max_requests_per_process) {
             ctx->procnode->diewhy = FCGID_DIE_LIFETIME_EXPIRED;
-            return_procnode(s, ctx->procnode,
-                            1 /* handled all requests */ );
-        } else
-            return_procnode(s, ctx->procnode,
-                            0 /* communication ok */ );
+            return_procnode(r, ctx->procnode, 1 /* handled all requests */ );
+        }
+        else
+            return_procnode(r, ctx->procnode, 0 /* communication ok */ );
 
         ctx->procnode = NULL;
     }
@@ -214,7 +207,7 @@ apr_status_t bucket_ctx_cleanup(void *thectx)
 static int getsfunc_fcgid_BRIGADE(char *buf, int len, void *arg)
 {
     apr_bucket_brigade *bb = (apr_bucket_brigade *) arg;
-    const char *dst_end = buf + len - 1;    /* leave room for terminating null */
+    const char *dst_end = buf + len - 1;        /* leave room for terminating null */
     char *dst = buf;
     apr_bucket *e = APR_BRIGADE_FIRST(bb);
     apr_status_t rv;
@@ -261,7 +254,8 @@ static int getsfunc_fcgid_BRIGADE(char *buf, int len, void *arg)
                 done = 1;
                 getColon = 0;
                 break;
-            } else if (getLF && (*src == ' ' || *src == '\t')) {
+            }
+            else if (getLF && (*src == ' ' || *src == '\t')) {
                 *dst++ = '\r';
                 *dst++ = '\n';
                 getLF = 0;
@@ -269,7 +263,8 @@ static int getsfunc_fcgid_BRIGADE(char *buf, int len, void *arg)
 
             if (*src == '\n') {
                 getLF = 1;
-            } else if (*src != '\r') {
+            }
+            else if (*src != '\r') {
                 *dst++ = *src;
             }
             src++;
@@ -327,20 +322,17 @@ handle_request(request_rec * r, int role, const char *argv0,
                 fcgi_request.cmdopts.ipc_comm_timeout;
 
             /* Apply a process slot */
-            bucket_ctx->procnode =
-                apply_free_procnode(r->server, &fcgi_request);
+            bucket_ctx->procnode = apply_free_procnode(r, &fcgi_request);
             if (bucket_ctx->procnode)
                 break;
 
             /* Avoid sleeping the very first time through if there are no
                busy processes; the problem is just that we haven't spawned
                anything yet, so waiting is pointless */
-            if (i > 0 || j > 0
-                || count_busy_processes(r->server, &fcgi_request)) {
+            if (i > 0 || j > 0 || count_busy_processes(r, &fcgi_request)) {
                 apr_sleep(apr_time_from_sec(1));
 
-                bucket_ctx->procnode =
-                    apply_free_procnode(r->server, &fcgi_request);
+                bucket_ctx->procnode = apply_free_procnode(r, &fcgi_request);
                 if (bucket_ctx->procnode)
                     break;
             }
@@ -355,10 +347,10 @@ handle_request(request_rec * r, int role, const char *argv0,
                                  &bucket_ctx->ipc) != APR_SUCCESS) {
                 proc_close_ipc(&bucket_ctx->ipc);
                 bucket_ctx->procnode->diewhy = FCGID_DIE_CONNECT_ERROR;
-                return_procnode(r->server, bucket_ctx->procnode,
-                                1 /* has error */ );
+                return_procnode(r, bucket_ctx->procnode, 1 /* has error */ );
                 bucket_ctx->procnode = NULL;
-            } else
+            }
+            else
                 break;
         }
     }
@@ -371,13 +363,11 @@ handle_request(request_rec * r, int role, const char *argv0,
     }
     bucket_ctx->active_time = bucket_ctx->procnode->last_active_time =
         apr_time_now();
+    bucket_ctx->procnode->diewhy = FCGID_DIE_KILLSELF;
 
     /* Write output_brigade to fastcgi server */
     if ((rv =
-         proc_write_ipc(&bucket_ctx->ipc,
-                        output_brigade)) != APR_SUCCESS) {
-        ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
-                      "mod_fcgid: error writing data to FastCGI server");
+         proc_write_ipc(&bucket_ctx->ipc, output_brigade)) != APR_SUCCESS) {
         bucket_ctx->has_error = 1;
         return HTTP_INTERNAL_SERVER_ERROR;
     }
@@ -414,7 +404,8 @@ handle_request(request_rec * r, int role, const char *argv0,
 
         ap_internal_redirect_handler(location, r);
         return HTTP_OK;
-    } else if (location && r->status == 200) {
+    }
+    else if (location && r->status == 200) {
         /* XX Note that if a script wants to produce its own Redirect 
          * body, it now has to explicitly *say* "Status: 302" 
          */
@@ -437,52 +428,30 @@ handle_request(request_rec * r, int role, const char *argv0,
     return cond_status;
 }
 
-int bridge_request(request_rec * r, int role, const char *argv0,
-                   fcgid_wrapper_conf * wrapper_conf)
+static int add_request_body(request_rec *r, apr_pool_t *request_pool,
+                            apr_bucket_brigade *output_brigade)
 {
-    apr_pool_t *request_pool = r->main ? r->main->pool : r->pool;
-    server_rec *s = r->server;
-    fcgid_server_conf *sconf = ap_get_module_config(s->module_config,
-                                                    &fcgid_module);
-    apr_status_t rv = APR_SUCCESS;
-    int seen_eos;
-    apr_off_t request_size = 0;
+    apr_bucket *bucket_input, *bucket_header;
     apr_file_t *fd = NULL;
-    int need_truncate = 1;
-    apr_off_t cur_pos = 0;
+    apr_off_t cur_pos = 0, request_size = 0;
+    apr_status_t rv;
     FCGI_Header *stdin_request_header;
-    apr_bucket_brigade *output_brigade;
-    apr_bucket *bucket_input, *bucket_header, *bucket_eos;
-    char **envp = ap_create_environment(request_pool,
-                                        r->subprocess_env);
-
-    /* Create brigade for the request to fastcgi server */
-    output_brigade =
-        apr_brigade_create(request_pool, r->connection->bucket_alloc);
-
-    /* Build the begin request and environ request, append them to output_brigade */
-    if (!build_begin_block
-        (role, r, r->connection->bucket_alloc, output_brigade)
-        || !build_env_block(r, envp, r->connection->bucket_alloc,
-                            output_brigade)) {
-        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-                      "mod_fcgid: can't build begin or env request");
-        return HTTP_INTERNAL_SERVER_ERROR;
-    }
+    fcgid_server_conf *sconf = ap_get_module_config(r->server->module_config,
+                                                    &fcgid_module);
+    int seen_eos;
 
     /* Stdin header and body */
-    /* XXX HACK: I have to read all the request into memory before sending it 
+    /* I have to read all the request into memory before sending it 
        to fastcgi application server, this prevents slow clients from 
        keeping the server in processing too long. 
        But sometimes it's not acceptable (think about uploading a large attachment)
-       file_bucket is a better choice in this case...
-       To do, or not to do, that is the question ^_^
+       Request will be stored in tmp file if the size larger than max_mem_request_len
      */
     seen_eos = 0;
     do {
-        apr_bucket_brigade *input_brigade =
-            apr_brigade_create(request_pool,
-                               r->connection->bucket_alloc);
+        apr_bucket_brigade *input_brigade = apr_brigade_create(request_pool,
+                                                               r->connection->
+                                                               bucket_alloc);
 
         if ((rv = ap_get_brigade(r->input_filters, input_brigade,
                                  AP_MODE_READBYTES,
@@ -532,10 +501,10 @@ int bridge_request(request_rec * r, int role, const char *argv0,
             request_size += len;
             if (request_size > sconf->max_request_len) {
                 ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-                              "mod_fcgid: HTTP request length %" APR_OFF_T_FMT 
-                              " (so far) exceeds MaxRequestLen (%" APR_OFF_T_FMT
-                              ")",
-                             request_size, sconf->max_request_len);
+                              "mod_fcgid: HTTP request length %" APR_OFF_T_FMT
+                              " (so far) exceeds MaxRequestLen (%"
+                              APR_OFF_T_FMT ")", request_size,
+                              sconf->max_request_len);
                 return HTTP_INTERNAL_SERVER_ERROR;
             }
 
@@ -543,11 +512,19 @@ int bridge_request(request_rec * r, int role, const char *argv0,
                 apr_size_t wrote_len;
                 static const char *fd_key = "fcgid_fd";
 
-                if (fd == NULL){
+                if (fd == NULL) {
                     void *tmp;
-                    apr_pool_userdata_get(&tmp, fd_key,
-                                          r->connection->pool);
+                    apr_pool_userdata_get(&tmp, fd_key, r->connection->pool);
                     fd = tmp;
+
+                    if (fd != NULL) {
+                        if ((rv = apr_file_trunc(fd, 0)) != APR_SUCCESS) {
+                            ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, r,
+                                          "mod_fcgid: can't truncate existing "
+                                          "temporary file");
+                            return HTTP_INTERNAL_SERVER_ERROR;
+                        }
+                    }
                 }
 
                 if (fd == NULL) {
@@ -574,12 +551,9 @@ int bridge_request(request_rec * r, int role, const char *argv0,
                     apr_pool_userdata_set((const void *) fd, fd_key,
                                           apr_pool_cleanup_null,
                                           r->connection->pool);
-                } else if (need_truncate) {
-                    need_truncate = 0;
-                    apr_file_trunc(fd, 0);
-                    cur_pos = 0;
                 }
-                // Write request to tmp file
+
+                /* Write request to tmp file */
                 if ((rv =
                      apr_file_write_full(fd, (const void *) data, len,
                                          &wrote_len)) != APR_SUCCESS
@@ -589,12 +563,13 @@ int bridge_request(request_rec * r, int role, const char *argv0,
                                   "mod_fcgid: can't write tmp file for stdin request");
                     return HTTP_INTERNAL_SERVER_ERROR;
                 }
-                // Create file bucket
+                /* Create file bucket */
                 bucket_stdin =
                     apr_bucket_file_create(fd, cur_pos, len, r->pool,
                                            r->connection->bucket_alloc);
                 cur_pos += len;
-            } else {
+            }
+            else {
                 if (APR_BUCKET_IS_HEAP(bucket_input))
                     apr_bucket_copy(bucket_input, &bucket_stdin);
                 else {
@@ -605,8 +580,7 @@ int bridge_request(request_rec * r, int role, const char *argv0,
                     bucket_stdin =
                         apr_bucket_heap_create(pcopydata, len,
                                                apr_bucket_free,
-                                               r->connection->
-                                               bucket_alloc);
+                                               r->connection->bucket_alloc);
                 }
             }
 
@@ -631,8 +605,7 @@ int bridge_request(request_rec * r, int role, const char *argv0,
     bucket_header =
         apr_bucket_heap_create((const char *) stdin_request_header,
                                sizeof(*stdin_request_header),
-                               apr_bucket_free,
-                               r->connection->bucket_alloc);
+                               apr_bucket_free, r->connection->bucket_alloc);
     if (!init_header(FCGI_STDIN, 1, 0, 0, stdin_request_header)) {
         ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                       "mod_fcgid: header overflow");
@@ -640,6 +613,40 @@ int bridge_request(request_rec * r, int role, const char *argv0,
     }
     APR_BRIGADE_INSERT_TAIL(output_brigade, bucket_header);
 
+    return 0;
+}
+
+int bridge_request(request_rec * r, int role, const char *argv0,
+                   fcgid_wrapper_conf * wrapper_conf)
+{
+    apr_pool_t *request_pool = r->main ? r->main->pool : r->pool;
+    apr_bucket_brigade *output_brigade;
+    apr_bucket *bucket_eos;
+    char **envp = ap_create_environment(request_pool,
+                                        r->subprocess_env);
+    int rc;
+
+    /* Create brigade for the request to fastcgi server */
+    output_brigade =
+        apr_brigade_create(request_pool, r->connection->bucket_alloc);
+
+    /* Build the begin request and environ request, append them to output_brigade */
+    if (!build_begin_block
+        (role, r, r->connection->bucket_alloc, output_brigade)
+        || !build_env_block(r, envp, r->connection->bucket_alloc,
+                            output_brigade)) {
+        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
+                      "mod_fcgid: can't build begin or env request");
+        return HTTP_INTERNAL_SERVER_ERROR;
+    }
+
+    if (role == FCGI_RESPONDER) {
+        rc = add_request_body(r, request_pool, output_brigade);
+        if (rc) {
+            return rc;
+        }
+    }
+
     /* The eos bucket now */
     bucket_eos = apr_bucket_eos_create(r->connection->bucket_alloc);
     APR_BRIGADE_INSERT_TAIL(output_brigade, bucket_eos);
