@@ -28,6 +28,10 @@
 #include "fcgid_global.h"
 #include "fcgid_conf.h"
 
+#ifndef DEFAULT_REL_RUNTIMEDIR /* Win32, etc. */
+#define DEFAULT_REL_RUNTIMEDIR "logs"
+#endif
+
 #define DEFAULT_IDLE_TIMEOUT 300
 #define DEFAULT_IDLE_SCAN_INTERVAL 120
 #define DEFAULT_BUSY_TIMEOUT 300
@@ -35,8 +39,8 @@
 #define DEFAULT_ERROR_SCAN_INTERVAL 3
 #define DEFAULT_ZOMBIE_SCAN_INTERVAL 3
 #define DEFAULT_PROC_LIFETIME (60*60)
-#define DEFAULT_SOCKET_PREFIX "logs/fcgidsock"
-#define DEFAULT_SHM_PATH "logs/fcgid_shm"
+#define DEFAULT_SOCKET_PREFIX DEFAULT_REL_RUNTIMEDIR "/fcgidsock"
+#define DEFAULT_SHM_PATH      DEFAULT_REL_RUNTIMEDIR "/fcgid_shm"
 #define DEFAULT_SPAWNSOCRE_UPLIMIT 10
 #define DEFAULT_SPAWN_SCORE 1
 #define DEFAULT_TERMINATION_SCORE 2
@@ -748,8 +752,11 @@ typedef struct {
     apr_size_t cur_id;
 } wrapper_id_info;
 
+/* FIXME thread safety issues when FcgidWrapper is used in .htaccess;
+ * see use of pconf
+ */
 const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
-                               const char *wrapperpath,
+                               const char *wrapper_cmdline,
                                const char *extension,
                                const char *virtual)
 {
@@ -760,13 +767,11 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
     wrapper_id_info *id_info;
     apr_size_t *wrapper_id;
     fcgid_wrapper_conf *wrapper = NULL;
+    apr_pool_t *wrapper_conf_pool = cmd->server->process->pconf; /* bad */
     fcgid_dir_conf *config = (fcgid_dir_conf *) dirconfig;
 
     /* Sanity checks */
 
-    if (wrapperpath == NULL)
-        return "Invalid wrapper file";
-
     if (virtual == NULL && extension != NULL && !strcasecmp(extension, WRAPPER_FLAG_VIRTUAL)) {
         virtual = WRAPPER_FLAG_VIRTUAL;
         extension = NULL;
@@ -798,21 +803,21 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
                               apr_pool_cleanup_null,
                               cmd->server->process->pool);
     }
-    /* Get wrapper_id for wrapperpath */
+    /* Get wrapper_id for wrapper_cmdline */
     if ((wrapper_id =
-         apr_hash_get(id_info->wrapper_id_hash, wrapperpath,
-                      strlen(wrapperpath))) == NULL) {
+         apr_hash_get(id_info->wrapper_id_hash, wrapper_cmdline,
+                      strlen(wrapper_cmdline))) == NULL) {
         wrapper_id =
             apr_pcalloc(cmd->server->process->pool, sizeof(*wrapper_id));
         *wrapper_id = id_info->cur_id++;
-        apr_hash_set(id_info->wrapper_id_hash, wrapperpath,
-                     strlen(wrapperpath), wrapper_id);
+        apr_hash_set(id_info->wrapper_id_hash, wrapper_cmdline,
+                     strlen(wrapper_cmdline), wrapper_id);
     }
 
-    wrapper = apr_pcalloc(cmd->server->process->pconf, sizeof(*wrapper));
+    wrapper = apr_pcalloc(wrapper_conf_pool, sizeof(*wrapper));
 
     /* Get wrapper path */
-    tmp = wrapperpath;
+    tmp = wrapper_cmdline;
     path = ap_getword_white(cmd->temp_pool, &tmp);
     if (path == NULL || *path == '\0')
         return "Invalid wrapper config";
@@ -823,7 +828,9 @@ const char *set_wrapper_config(cmd_parms * cmd, void *dirconfig,
         return missing_file_msg(cmd->pool, "Wrapper", path, rv);
     }
 
-    apr_cpystrn(wrapper->args, wrapperpath, _POSIX_PATH_MAX);
+    wrapper->exe = apr_pstrdup(wrapper_conf_pool, path);
+    /* FIXME no need to embed in structure (subject to correct pool usage) */
+    apr_cpystrn(wrapper->args, wrapper_cmdline, _POSIX_PATH_MAX);
     wrapper->inode = finfo.inode;
     wrapper->deviceid = finfo.device;
     wrapper->share_group_id = *wrapper_id;
@@ -867,7 +874,7 @@ fcgid_wrapper_conf *get_wrapper_info(const char *cgipath, request_rec * r)
     return NULL;
 }
 
-static int set_cmd_envvars(fcgid_cmd_options *cmdopts, apr_table_t *envvars)
+static int set_cmd_envvars(fcgid_cmd_env *cmdenv, apr_table_t *envvars)
 {
     const apr_array_header_t *envvars_arr;
     const apr_table_entry_t *envvars_entry;
@@ -885,17 +892,17 @@ static int set_cmd_envvars(fcgid_cmd_options *cmdopts, apr_table_t *envvars)
             if (envvars_entry[i].key == NULL
                 || envvars_entry[i].key[0] == '\0')
                 break;
-            apr_cpystrn(cmdopts->initenv_key[i], envvars_entry[i].key,
+            apr_cpystrn(cmdenv->initenv_key[i], envvars_entry[i].key,
                         INITENV_KEY_LEN);
-            apr_cpystrn(cmdopts->initenv_val[i], envvars_entry[i].val,
+            apr_cpystrn(cmdenv->initenv_val[i], envvars_entry[i].val,
                         INITENV_VAL_LEN);
         }
         if (i < INITENV_CNT) {
-            cmdopts->initenv_key[i][0] = '\0';
+            cmdenv->initenv_key[i][0] = '\0';
         }
     }
     else {
-        cmdopts->initenv_key[0][0] = '\0';
+        cmdenv->initenv_key[0][0] = '\0';
     }
 
     return overflow;
@@ -914,6 +921,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
     apr_status_t rv;
 
     cmdopts = apr_pcalloc(cmd->pool, sizeof *cmdopts);
+    cmdopts->cmdenv = apr_pcalloc(cmd->pool, sizeof *cmdopts->cmdenv);
 
     cmdopts->busy_timeout = DEFAULT_BUSY_TIMEOUT;
     cmdopts->idle_timeout = DEFAULT_IDLE_TIMEOUT;
@@ -943,9 +951,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
         const char *option = ap_getword_white(cmd->pool, &args);
         const char *val;
 
-        /* don't support BusyTimeout until BZ #47483 is fixed
-         * (The kludge in bucket_ctx_cleanup wouldn't have addressibility
-         * to the cmd-specific busy timeout setting.)
+        /* TODO: Consider supporting BusyTimeout.
          */
 
         if (!strcasecmp(option, "ConnectTimeout")) {
@@ -1038,7 +1044,7 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
                             option);
     }
 
-    if ((overflow = set_cmd_envvars(cmdopts, envvars)) != 0) {
+    if ((overflow = set_cmd_envvars(cmdopts->cmdenv, envvars)) != 0) {
         return apr_psprintf(cmd->pool, "mod_fcgid: environment variable table "
                             "overflow; increase INITENV_CNT in fcgid_pm.h from"
                             " %d to at least %d",
@@ -1051,7 +1057,8 @@ const char *set_cmd_options(cmd_parms *cmd, void *dummy, const char *args)
 }
 
 void get_cmd_options(request_rec *r, const char *cmdpath,
-                     fcgid_cmd_options *cmdopts)
+                     fcgid_cmd_options *cmdopts,
+                     fcgid_cmd_env *cmdenv)
 {
     fcgid_server_conf *sconf =
         ap_get_module_config(r->server->module_config, &fcgid_module);
@@ -1062,6 +1069,8 @@ void get_cmd_options(request_rec *r, const char *cmdpath,
 
     if (cmd_specific) { /* ignore request context configuration */
         *cmdopts = *cmd_specific;
+        *cmdenv = *cmdopts->cmdenv;
+        cmdopts->cmdenv = NULL;
         /* pick up configuration for values that can't be configured
          * on FcgidCmdOptions
          */
@@ -1078,7 +1087,7 @@ void get_cmd_options(request_rec *r, const char *cmdpath,
     cmdopts->min_class_process_count = sconf->min_class_process_count;
     cmdopts->proc_lifetime = sconf->proc_lifetime;
 
-    if ((overflow = set_cmd_envvars(cmdopts, sconf->default_init_env)) != 0) {
+    if ((overflow = set_cmd_envvars(cmdenv, sconf->default_init_env)) != 0) {
         ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                       "mod_fcgid: %d environment variables dropped; increase "
                       "INITENV_CNT in fcgid_pm.h from %d to at least %d",
@@ -1086,4 +1095,6 @@ void get_cmd_options(request_rec *r, const char *cmdpath,
                       INITENV_CNT,
                       INITENV_CNT + overflow);
     }
+
+    cmdopts->cmdenv = NULL;
 }
