@@ -23,6 +23,7 @@
 #include <sys/stat.h>
 #define CORE_PRIVATE
 #include "httpd.h"
+#include "apr_version.h"
 #include "apr_thread_proc.h"
 #include "apr_strings.h"
 #include "apr_portable.h"
@@ -45,8 +46,12 @@
 
 #if MODULE_MAGIC_NUMBER_MAJOR < 20081201
 #define ap_unixd_config unixd_config
-// #define ap_unixd_setup_child unixd_setup_child
-// #define ap_unixd_set_global_mutex_perms unixd_set_global_mutex_perms
+#endif
+
+#if APR_MAJOR_VERSION < 1
+#define APR_FPROT_UWRITE        APR_UWRITE
+#define APR_FPROT_UREAD         APR_UREAD
+#define APR_FPROT_UEXECUTE      APR_UEXECUTE
 #endif
 
 #define DEFAULT_FCGID_LISTENBACKLOG 5
@@ -56,7 +61,6 @@ typedef struct {
 } fcgid_namedpipe_handle;
 
 static int g_process_counter = 0;
-static apr_pool_t *g_inode_cginame_map = NULL;
 
 static apr_status_t ap_unix_create_privileged_process(apr_proc_t *newproc,
                                                       const char *progname,
@@ -97,10 +101,8 @@ static apr_status_t ap_unix_create_privileged_process(apr_proc_t *newproc,
     }
 
     i = 0;
-    if (args) {
-        while (args[i]) {
-            i++;
-        }
+    while (args[i]) {
+        i++;
     }
     /* allocate space for 4 new args, the input args, and a null terminator */
     newargs = apr_palloc(p, sizeof(char *) * (i + 4));
@@ -174,12 +176,11 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
                                                     &fcgid_module);
     apr_status_t rv = APR_SUCCESS;
     apr_file_t *file;
+    apr_proc_t tmpproc;
     int omask, retcode, unix_socket;
     char **proc_environ;
     struct sockaddr_un unix_addr;
     apr_procattr_t *procattr = NULL;
-    char key_name[_POSIX_PATH_MAX];
-    void *dummy;
     int argc;
     const char *wargv[APACHE_ARG_MAX];
     const char *word; /* For wrapper */
@@ -199,18 +200,6 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
     }
     wargv[argc] = NULL;
 
-    /* Initialize the variables */
-    if (!g_inode_cginame_map) {
-        rv = apr_pool_create(&g_inode_cginame_map,
-                             procinfo->main_server->process->pconf);
-        if (rv != APR_SUCCESS) {
-            ap_log_error(APLOG_MARK, APLOG_WARNING, rv,
-                         procinfo->main_server,
-                         "mod_fcgid: can't create CGI name map table");
-            return APR_ENOMEM;
-        }
-    }
-
     /* 
        Create UNIX domain socket before spawn 
      */
@@ -224,6 +213,9 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
                  getpid(), g_process_counter++);
     strncpy(procnode->socket_path, unix_addr.sun_path,
             sizeof(procnode->socket_path) - 1);
+    strncpy(procnode->executable_path,
+            (lpszwapper != NULL && lpszwapper[0] != '\0') ? wargv[0] : procinfo->cgipath,
+            sizeof(procnode->executable_path) - 1);
 
     /* Unlink the file just in case */
     unlink(unix_addr.sun_path);
@@ -259,6 +251,17 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
         return errno;
     }
 
+    /* IPC directory permissions are safe, but avoid confusion */
+    /* Not all flavors of unix use the current umask for AF_UNIX perms */
+
+    rv = apr_file_perms_set(unix_addr.sun_path, APR_FPROT_UREAD|APR_FPROT_UWRITE|APR_FPROT_UEXECUTE);
+    if (rv != APR_SUCCESS) {
+        ap_log_error(APLOG_MARK, APLOG_CRIT, rv, main_server,
+                     "mod_fcgid: Couldn't set permissions on unix domain socket %s",
+                     unix_addr.sun_path);
+        return rv;
+    }
+
     /* Listen the socket */
     if (listen(unix_socket, DEFAULT_FCGID_LISTENBACKLOG) < 0) {
         ap_log_error(APLOG_MARK, APLOG_ERR, errno, main_server,
@@ -311,8 +314,6 @@ proc_spawn_process(char *lpszwapper, fcgid_proc_info * procinfo,
     }
 
     /* Prepare the fork */
-    procnode->proc_id = apr_pcalloc(procnode->proc_pool, 
-                                    sizeof(apr_proc_t));
     if (
    (rv =
     apr_procattr_create(&procattr, procnode->proc_pool)) != APR_SUCCESS
@@ -345,12 +346,16 @@ APR_SUCCESS
     }
 
     /* fork and exec now */
+    /* Note, don't pass &(procnode->proc_id) to fcgid_create_privileged_process(),
+     * for it's a share memory address, both parent and child process may modify 
+     * procnode->proc_id->pid, so sometimes it's 0 and sometimes it's >0
+     */
     if (lpszwapper != NULL && lpszwapper[0] != '\0') {
-        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, procinfo->main_server,
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, procinfo->main_server,
                      "mod_fcgid: call %s with wrapper %s",
                      procinfo->cgipath, lpszwapper);
         if ((rv =
-             fcgid_create_privileged_process(procnode->proc_id,
+             fcgid_create_privileged_process(&tmpproc,
                                              wargv[0], wargv,
                                              (const char *const *)
                                              proc_environ, procattr,
@@ -361,13 +366,14 @@ APR_SUCCESS
                          "mod_fcgid: can't create wrapper process for %s",
                          procinfo->cgipath);
             close(unix_socket);
+            procnode->proc_id = tmpproc;
             return rv;
         }
     } else {
         argv[0] = procinfo->cgipath;
         argv[1] = NULL;
         if ((rv =
-             fcgid_create_privileged_process(procnode->proc_id,
+             fcgid_create_privileged_process(&tmpproc,
                                              procinfo->cgipath,
                                              argv,
                                              (const char *const *)
@@ -378,31 +384,14 @@ APR_SUCCESS
             ap_log_error(APLOG_MARK, APLOG_ERR, rv, procinfo->main_server,
                          "mod_fcgid: can't create process");
             close(unix_socket);
+            procnode->proc_id = tmpproc;
             return rv;
         }
     }
 
-    /* Set the (deviceid, inode) -> fastcgi path map for log */
-    apr_snprintf(key_name, _POSIX_PATH_MAX, "%lX%lX",
-                 (unsigned long) procnode->inode,
-                 (unsigned long) procnode->deviceid);
-    dummy = NULL;
-    apr_pool_userdata_get(&dummy, key_name, g_inode_cginame_map);
-    if (!dummy) {
-        /* Insert a new item if key not found */
-        const char *put_key = apr_psprintf(g_inode_cginame_map, "%lX%lX",
-                                           (unsigned long) procnode->inode,
-                                           (unsigned long) procnode->deviceid);
-        const char *fcgipath = apr_psprintf(g_inode_cginame_map, "%s",
-                                            procinfo->cgipath);
-
-        if (put_key && fcgipath)
-            apr_pool_userdata_set(fcgipath, put_key, NULL,
-                                  g_inode_cginame_map);
-    }
-
     /* Close socket before try to connect to it */
     close(unix_socket);
+    procnode->proc_id = tmpproc;
 
     return APR_SUCCESS;
 }
@@ -420,7 +409,7 @@ proc_kill_gracefully(fcgid_procnode * procnode, server_rec * main_server)
         kill(getpid(), SIGTERM);
         return APR_EACCES;
     }
-    rv = apr_proc_kill(procnode->proc_id, SIGTERM);
+    rv = apr_proc_kill(&(procnode->proc_id), SIGTERM);
     if (ap_unixd_config.suexec_enabled && seteuid(ap_unixd_config.user_id) != 0)
     {
         kill(getpid(), SIGTERM);
@@ -440,7 +429,7 @@ apr_status_t proc_kill_force(fcgid_procnode * procnode,
         kill(getpid(), SIGTERM);
         return APR_EACCES;
     }
-    rv = apr_proc_kill(procnode->proc_id, SIGKILL);
+    rv = apr_proc_kill(&(procnode->proc_id), SIGKILL);
     if (ap_unixd_config.suexec_enabled && seteuid(ap_unixd_config.user_id) != 0)
     {
         kill(getpid(), SIGTERM);
@@ -456,7 +445,7 @@ proc_wait_process(server_rec * main_server, fcgid_procnode * procnode)
     int exitcode;
     apr_exit_why_e exitwhy;
 
-    rv = apr_proc_wait(procnode->proc_id, &exitcode, &exitwhy, APR_NOWAIT);
+    rv = apr_proc_wait(&(procnode->proc_id), &exitcode, &exitwhy, APR_NOWAIT);
     if (rv == APR_CHILD_DONE || rv == APR_EGENERAL) {
         /* Log why and how it die */
         proc_print_exit_info(procnode, exitcode, exitwhy, main_server);
@@ -645,6 +634,7 @@ static apr_status_t socket_writev(fcgid_ipc * ipc_handle,
                                   struct iovec *vec, int nvec,
                                   int *writecnt)
 {
+    apr_status_t rv;
     int retcode, unix_socket;
     fcgid_namedpipe_handle *handle_info;
     struct pollfd pollfds[1];
@@ -659,34 +649,69 @@ static apr_status_t socket_writev(fcgid_ipc * ipc_handle,
             return APR_SUCCESS;
         }
     } while (retcode == -1 && APR_STATUS_IS_EINTR(errno));
-    if (!APR_STATUS_IS_EAGAIN(errno))
-        return errno;
+    rv = errno;
 
-    /* poll() */
-    pollfds[0].fd = unix_socket;
-    pollfds[0].events = POLLOUT;
-    do {
-        retcode = poll(pollfds, 1, ipc_handle->communation_timeout * 1000);
-    } while (retcode <= 0 && APR_STATUS_IS_EINTR(errno));
-    if (retcode == -1)
-        return errno;
+    if (APR_STATUS_IS_EAGAIN(rv)) {
+        /* poll() */
+        pollfds[0].fd = unix_socket;
+        pollfds[0].events = POLLOUT;
+        do {
+            retcode = poll(pollfds, 1, ipc_handle->communation_timeout * 1000);
+        } while (retcode < 0 && APR_STATUS_IS_EINTR(errno));
 
-    /* Write again */
-    do {
-        if ((retcode = writev(unix_socket, vec, nvec)) > 0) {
-            *writecnt = retcode;
-            return APR_SUCCESS;
+        if (retcode < 0) {
+            rv = errno;
+        }
+        else if (retcode == 0) {
+            rv = APR_TIMEUP;
+        }
+        else {
+            /* Write again */
+            do {
+                if ((retcode = writev(unix_socket, vec, nvec)) > 0) {
+                    *writecnt = retcode;
+                    return APR_SUCCESS;
+                }
+            } while (retcode == -1 && APR_STATUS_IS_EINTR(errno));
+            rv = errno;
         }
-    } while (retcode == -1 && APR_STATUS_IS_EINTR(errno));
-
-    if (retcode == -1) {
-        ap_log_rerror(APLOG_MARK, APLOG_INFO, apr_get_os_error(),
-                      ipc_handle->request,
-                      "mod_fcgid: error writing data, FastCGI server closed connection");
-        return APR_EPIPE;
     }
 
-    return errno;
+    if (APR_STATUS_IS_EAGAIN(rv)) {
+        /* socket is writable, but we can't write the entire buffer; try to write a
+         * smaller amount, and if even that fails then sleep
+         */
+        size_t to_write = vec[0].iov_len;
+        int slept = 0;
+        const apr_interval_time_t sleep_time = APR_USEC_PER_SEC / 4;
+        const int max_sleeps = 8;
+
+        do {
+            if ((retcode = write(unix_socket, vec[0].iov_base, to_write)) > 0) {
+                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, ipc_handle->request,
+                              "wrote %d byte(s) and slept %d time(s) after EAGAIN",
+                              retcode, slept);
+                *writecnt = retcode;
+                return APR_SUCCESS;
+            }
+            if (APR_STATUS_IS_EAGAIN(errno)) {
+                if (to_write == 1) {
+                    apr_sleep(sleep_time);
+                    ++slept;
+                }
+                else {
+                    to_write /= 2;
+                }
+            }
+        } while ((APR_STATUS_IS_EINTR(errno) || APR_STATUS_IS_EAGAIN(errno))
+                 && slept < max_sleeps);
+        rv = errno;
+    }
+
+    ap_log_rerror(APLOG_MARK, APLOG_INFO, rv,
+                  ipc_handle->request,
+                  "mod_fcgid: error writing data to FastCGI server");
+    return rv;
 }
 
 static apr_status_t writev_it_all(fcgid_ipc * ipc_handle,
@@ -747,8 +772,11 @@ apr_status_t proc_write_ipc(fcgid_ipc * ipc_handle,
         apr_size_t len;
         const char* base;
         if ((rv = apr_bucket_read(e, &base, &len,
-                                  APR_BLOCK_READ)) != APR_SUCCESS)
+                                  APR_BLOCK_READ)) != APR_SUCCESS) {
+            ap_log_rerror(APLOG_MARK, APLOG_WARNING, rv, ipc_handle->request,
+                          "mod_fcgid: can't read request from bucket");
             return rv;
+        }
 
         vec[nvec].iov_len = len;
         vec[nvec].iov_base = (char*) base;
@@ -776,23 +804,13 @@ void
 proc_print_exit_info(fcgid_procnode * procnode, int exitcode,
                      apr_exit_why_e exitwhy, server_rec * main_server)
 {
-    const char *cgipath = NULL;
     const char *diewhy = NULL;
     char signal_info[HUGE_STRING_LEN];
-    char key_name[_POSIX_PATH_MAX];
     int signum = exitcode;
-    void* tmp;
+    int loglevel = APLOG_INFO;
 
     memset(signal_info, 0, HUGE_STRING_LEN);
 
-    /* Get the file name infomation base on inode and deviceid */
-    apr_snprintf(key_name, _POSIX_PATH_MAX, "%lX%lX",
-                 (unsigned long) procnode->inode,
-                 (unsigned long) procnode->deviceid);
-    apr_pool_userdata_get(&tmp, key_name,
-                          g_inode_cginame_map);
-    cgipath = tmp;
-
     /* Reasons to exit */
     switch (procnode->diewhy) {
     case FCGID_DIE_KILLSELF:
@@ -817,7 +835,8 @@ proc_print_exit_info(fcgid_procnode * procnode, int exitcode,
         diewhy = "shutting down";
         break;
     default:
-        diewhy = "unknow";
+        loglevel = APLOG_ERR;
+        diewhy = "unknown";
     }
 
     /* Get signal info */
@@ -832,6 +851,7 @@ proc_print_exit_info(fcgid_procnode * procnode, int exitcode,
             break;
 
         default:
+            loglevel = APLOG_ERR;
             if (APR_PROC_CHECK_CORE_DUMP(exitwhy)) {
                 apr_snprintf(signal_info, HUGE_STRING_LEN - 1,
                              "get signal %d, possible coredump generated",
@@ -850,12 +870,7 @@ proc_print_exit_info(fcgid_procnode * procnode, int exitcode,
     }
 
     /* Print log now */
-    if (cgipath)
-        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, main_server,
-                     "mod_fcgid: process %s(%" APR_PID_T_FMT ") exit(%s), %s",
-                     cgipath, procnode->proc_id->pid, diewhy, signal_info);
-    else
-        ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, main_server,
-                     "mod_fcgid: can't get CGI name while exiting, exitcode: %d",
-                     exitcode);
+    ap_log_error(APLOG_MARK, loglevel, 0, main_server,
+                 "mod_fcgid: process %s(%" APR_PID_T_FMT ") exit(%s), %s",
+                 procnode->executable_path, procnode->proc_id.pid, diewhy, signal_info);
 }
